type bstmt =
    Lit of int     (* Push an integer literal *)
  | Slit of string (* Push a string literal *)
  | Blit of bool   (* Push a boolean literal *)
  | Alit of int    (* Push an array literal. This will indicate size of array *)
  | Drp            (* Discard a value *)
  | Arpt           (* Repeat *)
  | Bin of Ast.op  (* Perform arithmetic on top of stack *)
  | Loda of int    (* Fetch global array variable using index on top of stack *)
  | Lod of int     (* Fetch global variable *)
  | Str of int     (* Store global variable *)
  | Stra of int    (* Store global array variable using index on top of stack *)
  | Lfp of int     (* Load frame pointer relative *)
  | Lfpa of int    (* Load frame pointer relative with array index *)
  | Sfp of int     (* Store frame pointer relative *)
  | Sfpa of int    (* Store frame pointer relative with array index *)
  | Jsr of int     (* Call function by absolute address *)
  | Ent of int     (* Push FP, FP -> SP, SP += i *)
  | Rts of int     (* Restore FP, SP, consume formals, push result *)
  | Beq of int     (* Branch relative if top-of-stack is false *)
  | Bne of int     (* Branch relative if top-of-stack is true *)
  | Bra of int     (* Branch relative *)
  | Hlt            (* Terminate *)

type prog = {
    num_globals : int;   (* Number of global variables *)
    text : bstmt array; (* Code for all the program *)
  }

let string_of_stmt = function
    Lit(i) -> "Lit " ^ string_of_int i
  | Slit(s) -> "Slit \"" ^ s ^ "\""
  | Blit(b) -> "Blit " ^ string_of_bool b
  | Alit(l) -> "Alit " ^ string_of_int l
  | Drp -> "Drp"
  | Arpt -> "Arpt"
  | Bin(Ast.Add) -> "Add"
  | Bin(Ast.Sub) -> "Sub"
  | Bin(Ast.Mult) -> "Mul"
  | Bin(Ast.Div) -> "Div"
  | Bin(Ast.Equal) -> "Eql"
  | Bin(Ast.Neq) -> "Neq"
  | Bin(Ast.Less) -> "Lt"
  | Bin(Ast.Leq) -> "Leq"
  | Bin(Ast.Greater) -> "Gt"
  | Bin(Ast.Geq) -> "Geq"
  | Bin(Ast.Neg) -> "Neg"
  | Bin(Ast.Mod) -> "Mod"
  | Bin(Ast.Concat) -> "Concat"
  | Bin(Ast.Or) -> "Or"
  | Bin(Ast.And) -> "And"
  | Lod(i) -> "Lod " ^ string_of_int i
  | Loda(i) -> "Loda " ^ string_of_int i 
  | Str(i) -> "Str " ^ string_of_int i
  | Stra(i) -> "Stra " ^ string_of_int i
  | Lfp(i) -> "Lfp " ^ string_of_int i
  | Lfpa(i) -> "Lfpa " ^ string_of_int i
  | Sfp(i) -> "Sfp " ^ string_of_int i
  | Sfpa(i) -> "Sfpa " ^ string_of_int i
  | Jsr(i) -> "Jsr " ^ string_of_int i
  | Ent(i) -> "Ent " ^ string_of_int i
  | Rts(i) -> "Rts " ^ string_of_int i
  | Bne(i) -> "Bne " ^ string_of_int i
  | Beq(i) -> "Beq " ^ string_of_int i
  | Bra(i) -> "Bra " ^ string_of_int i
  | Hlt    -> "Hlt"

let string_of_prog p =
  string_of_int p.num_globals ^ " global variables\n" ^
  let funca = Array.mapi
      (fun i s -> string_of_int i ^ " " ^ string_of_stmt s) p.text
  in String.concat "\n" (Array.to_list funca)

