open Ast
open Bytecode

module StringMap = Map.Make(String)

(* Custom compilation exception *)
exception CompilationFailure of string

(* Symbol table: Information about all the names in scope *)
type env = {
    question_index : int StringMap.t; (* Index for each function *)
    mutable global_index   : (int * t) StringMap.t; (* "Address" for global variables *)
		function_index : (int * int * t) StringMap.t; (* "Address" for functions *)
    mutable local_index    : (int * t) StringMap.t; (* FP offset for args, locals *)
  }
	
type scope = 
	| ScGlobal
	| ScLocal
	
(* Find variable and return 3-tuple with index, type, and scope where it was found *)
let find_variable n env = 
	let ((i, t), s) = try 
		(StringMap.find n env.local_index, ScLocal)
		with Not_found -> try 
			(StringMap.find n env.global_index, ScGlobal)
			with Not_found -> raise (Failure ("undeclared variable " ^ n)) in
		(i, t, s)
		
(* Find question and return its index *)
let find_question n env = 
	let i = try
		(StringMap.find n env.question_index)
		with Not_found -> raise (Failure ("undeclared question " ^ n)) in i
		
(* Find function and return its index *)
let find_function n env =
	let i = try
		(StringMap.find n env.function_index)
		with Not_found -> raise (Failure ("undeclared function " ^ n)) in i

(* Declare a variable type *)		
let decl_variable_type s t env =
	let (i, st, sc) = find_variable s env in
		(match st with
			| Undefined -> 
				(match sc with
					| ScLocal -> 
						env.local_index <- StringMap.add s (i, t) env.local_index
					| ScGlobal ->
						env.global_index <- StringMap.add s (i, t) env.global_index)
			| _ ->
				raise (Failure ("cannot re-declare \"" ^ s ^ "\"")))

(* Create list of 3-tuple with address, identifier, and type *)
let rec enum stride n = function
	| [] -> []
	| hd::tl -> match hd with
							| VarDecl(i,e) -> (n, i, Undefined) :: enum stride (n+stride) tl
							| ArrDecl(e1, i, e2) -> (n, i, Undefined) :: enum stride (n+stride) tl

(* Create list of 2-tuple with address and identifier *)
let rec qenum stride n = function
	| [] -> []
	| hd::tl -> (n, hd) :: qenum stride (n+stride) tl

(* Create list of 4-tuple with address, identifier, number of formals, and type *)
let rec fenum stride n = function
	| [] -> []
	| (id, cnt)::tl -> (n, id, cnt, Undefined) :: fenum stride (n+stride) tl	

(* Create a map where key is the identifier and value is address *)
let string_map_questions map pairs =
  List.fold_left (fun m (i, n) -> if StringMap.mem n m then
						raise (Failure ("question already declared: " ^ n)) else
						StringMap.add n i m) map pairs

(* Create a map where key is the identifier and value is a tuple of address and type *)
let string_map_variables map list =
  List.fold_left (fun m (i, n, t) -> if StringMap.mem n m then 
				  	raise (Failure ("variable already declared: " ^ n)) else
				  	StringMap.add n (i, t) m) map list	
						
(* Create a map where key is the identifier and value is a 3-tuple of addres, type, and number of formals *)
let string_map_functions map list =
	List.fold_left (fun m (i, n, c, t) -> if StringMap.mem n m then
						raise (Failure ("function already declared: " ^ n)) else
						StringMap.add n (i, c, t) m) map list

(* Raise failure if types are not equal. *)
(* If type is undefined, type checking will be performed at run time. *)
let require_type t t' = 
	if t <> Undefined && t' <> Undefined then
		if t <> t' then
			raise (Failure ("expected " ^ string_of_type t ^ " but received " ^ string_of_type t'))
		else
			()
	else
		()
		
(* Match argument with a comparable type, otherwise raise a failure. *)
(* If type is undefined, type checking will be performed at run time *)
let require_comparable c = match c with
	| Int | Boolean | String | Undefined -> ()
	| _ -> raise (Failure ("expected Int, Boolean, or String but received " ^ string_of_type c)) 

(* match argument with array and return array type, otherwise raise a failure *)
(* If type is undefined, type checking will be performed at run time *)
let get_array_type a = match a with
	| Array(t) -> t
	| Undefined -> Undefined
	| _ -> raise (Failure ("expected array but received " ^ string_of_type a)) 

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or question *)
let translate (globals, functions, questions, execute) =
	(* Allocate addresses for global variables *)
	let builtin_globals = [VarDecl("correct", BLiteral(false));
												 VarDecl("askCount", Literal(0));
												 VarDecl("correctCount", Literal(0))] in
	let globals = {globals with decls = builtin_globals @ globals.decls} in
	let global_index = string_map_variables StringMap.empty (enum 1 0 globals.decls) in 
	let global_var_count = List.length globals.decls in
	
	(* Allocate address for functions *)
	let function_list = fenum 1 0 (List.map (fun f -> (f.name, List.length f.formals)) functions) in
	let function_index = string_map_functions StringMap.empty function_list in
	
	(* Allocate addresses for questions *)
	let question_index = string_map_questions StringMap.empty 
		(qenum 1 (List.length function_list) (List.map (fun q -> q.name) questions)) in
	
	let translate env decl =
		(* Bookkeeping: FP offsets for locals *)
		let num_locals = List.length decl.decls
		and num_formals = List.length decl.formals
		and local_offsets = (enum 1 1 decl.decls)
		and formal_offsets = (enum (-1) (-2) (List.map (fun f -> VarDecl(f, Noexpr)) decl.formals) ) in
		env.local_index <- string_map_variables StringMap.empty (formal_offsets @ local_offsets);
		
		let rec expr = function
			| Literal i -> ([Lit i], Int)
			| SLiteral s -> ([Slit s], String)
			| BLiteral b -> ([Blit b], Boolean)
			| ALiteral e ->
					(match e with
						| [] ->  raise (Failure ("array may not be empty"))
						| hd :: tl -> 
								let rec evlarray ls t = function
									| [] -> List.rev ls
									| hd :: tl -> 
											let (hd', hdt') = expr hd in
												if hdt' <> t then
													raise (Failure ("array may only contain elements of type " ^ string_of_type t))
												else
													evlarray ((List.rev hd') @ ls) t tl in											
									let (hd', hdt') = expr hd in
										(hd' @ (evlarray [] hdt' tl) @ [Alit(List.length e)], Array(hdt')))
			| ALitRpt(e1, e2) ->
					let (e1', t1) = expr e1 in
						require_type Int t1;
						let (e2', t2) = expr e2 in
							(e2' @ e1' @ [Arpt], Array(t2))
			| Id s ->
					let (i, t, sc) = find_variable s env in
						(match sc with
							| ScLocal  -> ([Lfp i], t) 
							| ScGlobal -> ([Lod i], t)) 
			| ArrId(s, e) ->
					let (e', et') = expr e in
						require_type Int et';
						let (i, st, sc) = find_variable s env in
							let st' = get_array_type st in
								(match sc with
									| ScLocal  -> (e' @ [Lfpa i], st')
									| ScGlobal -> (e' @ [Loda i], st'))
			| Binop (e1, op, e2) -> 
					let (e1', t1) = expr e1 and (e2', t2) = expr e2 in
						let rt = match op with
											| Add | Sub | Mult | Div | Mod ->
													require_type Int t1; require_type Int t2; Int
											| Less | Leq | Greater | Geq ->
													require_type Int t1; require_type Int t2; Boolean
											| Equal | Neq ->
													require_type t1 t2; require_comparable t1; require_comparable t2; Boolean
											| And | Or ->
													require_type Boolean t1; require_type Boolean t2; Boolean
											| Concat ->
													require_comparable t1; require_comparable t2; String
											| Neg ->
													require_type Boolean t2; Boolean in
						(e1' @ e2' @ [Bin op], rt) 
			| Assign (s, e1, e2) -> 
					let (e2', t2) = expr e2 in
						let(i, st, sc) = find_variable s env in
						(e2' @ (match e1 with
							| Noexpr ->
									if st = t2 || st = Undefined || t2 = Undefined then
										(match sc with
											| ScLocal  -> [Sfp i]
											| ScGlobal -> [Str i])
									else
										raise (Failure ("cannot assign type " ^ string_of_type t2 ^ " to " ^ s ^ " of type " ^ string_of_type st))
							| _ ->
									let (e1', t1) = expr e1 in
										require_type Int t1;
										let st' = get_array_type st in 
											if st' = t2 || st = Undefined || t2 = Undefined  then
												e1' @
												(match sc with
													| ScLocal  -> [Sfpa i]
													| ScGlobal -> [Stra i])
											else
												raise (Failure ("cannot assign type " ^ string_of_type t2 ^ " to " ^ s ^ " of type " ^ string_of_type st'))
							), Void)
			| Que(qlist) ->
					(match decl.t with
						| Execute -> 
								let rec que ls = function
									| [] -> ls
									| hd :: tl -> 
											let i = find_question hd env in
												que (Jsr(i) :: ls) tl in
								(que [] qlist, Void)
						| _ -> 
								raise (Failure ("cannot define question transition outside of execute block")))
			| Call(id, el) ->
					let elc = List.length el in
						let (i, c, t) = find_function id env in
							if c = elc then
								let rec push_args ls = function
									| [] -> ls
									| hd::tl -> 
											let (e, _) = expr hd in
												push_args (e @ ls) tl in
								(push_args [] el @ [Jsr i ], Undefined)
							else
								raise( Failure("call to " ^ id ^ " provides " ^ 
									string_of_int elc ^ " but " ^ string_of_int c ^ " were expected"))	  				
			| Noexpr -> ([], Void) 
			
		in let rec stmt ps = try (match ps with
				| Block sl ->  List.concat (List.map stmt sl)
				| Expr e -> fst (expr e) @ [Drp]
				| If (p, t, f) -> let t' = stmt t and f' = stmt f in
					  let (p', pt) = expr p in require_type Boolean pt;
						p' @ [Beq(2 + List.length t')] @
						t' @ [Bra(1 + List.length f')] @ f'
				| Repeat (e1, e2, b) ->
		  			let b' = stmt b and (e1', t1) = expr e1 and (e2', _) = expr e2 in
							require_type Boolean(t1); let be2' = b' @ e2' in
		  					[Bra (1 + List.length be2')] @ be2' @ e1' @
		  					[Bne (-(List.length be2' + List.length e1'))]
				| Decls dl ->
						let decl = function
							| VarDecl(s, e) ->
									let (_, t1) = expr e in
										decl_variable_type s t1 env;
										fst (expr (Assign(s, Noexpr, e)))
							| ArrDecl(e1, s, e2) ->
										let(e2', t2) = expr e2 in 
											decl_variable_type s (Array(t2)) env;
											fst (expr (Assign(s, Noexpr, ALitRpt(e1, e2)))) in
						List.concat (List.map decl dl)
				| Return e -> 
						(match decl.t with
							| Function ->  fst (expr e) @ [Rts num_formals]
							| _ -> raise ( Failure("cannot call return outside of function"))))					 
			with Failure(s) -> raise(CompilationFailure("\n" ^ s ^ "\nin: " ^ Ast.string_of_stmt ps))		
			
			in match decl.t with
			  | Question | Execute | Function ->
						let sd = stmt (Decls decl.decls) in
							let bd = stmt (Block decl.body) in
								[Ent num_locals] @ sd @ bd @ (match decl.t with
									| Question -> [Jsr(-1); Rts 0]
									| Execute -> [Jsr(-2)]
									| Function -> [Rts num_formals]
									| _ -> [])                     
				| Global -> stmt (Block decl.body)
			
		in let env = { question_index = question_index; 
									 global_index = global_index;
									 function_index = function_index;
									 local_index = StringMap.empty } in
									
		(* Code to start the program *)
		
		(* First, initialize global variables *)
		let globals = {globals with body = [Decls globals.decls]; decls = [];} in
		let global_decls = translate env globals in
		
		(* Second, compile the functions *)
		let function_bodies = List.map (fun f -> translate env f) functions in
		
		(* Third, execute the Execute block*)
		let execute_body = (translate env execute) @ [Hlt] in
		
		(* Fourth, compile the questions *)
		let builtin_question_locals = [VarDecl("prompt", SLiteral(""));
																	 VarDecl("choice", ALiteral([SLiteral("")]));
																	 VarDecl("answer", ALiteral([SLiteral("")]))] in
		let question_bodies = List.map (fun q ->
																			let q = {q with decls = builtin_question_locals @ q.decls} in
																				translate env q) questions in
		
		(* Calculate question and function entry points by adding their lengths *)
		let offset = List.length global_decls + List.length execute_body in
		let (que_func_offset_list, _) = List.fold_left
			(fun (l,i) f -> (i :: l, (i + List.length f))) ([],0) (function_bodies @ question_bodies) in
		let que_func_offset = Array.of_list (List.rev que_func_offset_list) in
		
		{ num_globals = global_var_count;
    	(* Concatenate the compiled questions and replace the question
      	indexes in Jsr statements with PC values *)
    	text = Array.of_list (List.map (function
				Jsr i when i > (-1) -> Jsr (que_func_offset.(i) + offset)
      | _ as s -> s) (global_decls @ execute_body @ (List.concat function_bodies) @ (List.concat question_bodies)))
  	}
