open Ast
open Bytecode

(* Custom execute exception *)
exception ExecuteFailure of string

(* Data types that may be stored in the stack and globals *)
type data = 
	| Integer of int
	| Boolean of bool
	| String  of string
	| Array   of data array
	| Undefined

(* return integer from data, otherwise raise failure *)
let int_of_data = function
	| Integer i -> i
	| _ -> raise(Failure("expected int"))

(* return boolean from data, otherwise raise failure *)
let bool_of_data = function
	| Boolean b -> b
	| _ -> raise(Failure("expected boolean"))

(* return string from data, otherwise raise failure *)
let sstring_of_data = function
	| String s -> s
	| _ -> raise(Failure("expected string"))

(* return array from data, otherwise raise failure *)
let array_of_data = function
	| Array a -> a
	| _ -> raise(Failure("expected array"))

(* require t1 and t2 to be of same data type. otherwise, raise failure *)
let require_same_type t1 t2 =
	(match t1 with
		| Integer x -> (match t2 with Integer y -> () | _ -> raise(Failure("expected int")))
		| Boolean x -> (match t2 with Boolean y -> () | _ -> raise(Failure("expected boolean")))
		| String x  -> (match t2 with String y  -> () | _ -> raise(Failure("expected string")))
		| Array x   -> (match t2 with Array y  ->  () | _ -> raise(Failure("expected array")))
		| Undefined -> ())	 

(* Data type to string *)
let rec string_of_data = function
	| Integer i -> string_of_int i
	| Boolean b -> string_of_bool b
	| String s  -> s
	| Array a   -> "{" ^ String.concat ";" (Array.to_list (Array.map string_of_data a)) ^ "}"
	| Undefined -> ""

(* Stack layout just after "Ent":

              <-- SP
   Local n
   ...
   Local 0
   Saved FP   <-- FP
   Saved PC
   Arg 0
   ...
   Arg n *)

let execute_prog prog =
  let stack = Array.make 1024 (Undefined)
  and globals = Array.make prog.num_globals (Undefined) in
	
  (* Set all elements in stack from f to l to Undefined *)
  let clear f l =
	Array.fill stack f l Undefined in

  let rec exec fp sp pc = try (match prog.text.(pc) with
	  | Lit i  -> stack.(sp) <- Integer i ; exec fp (sp+1) (pc+1)
	  | Blit b -> stack.(sp) <- Boolean b ; exec fp (sp+1) (pc+1)
	  | Slit s -> stack.(sp) <- String s ; exec fp (sp+1) (pc+1)
	  | Alit i -> stack.(sp-i) <- Array(Array.sub stack (sp - i) i) ; exec fp (sp-i+1) (pc+1)
	  | Arpt   -> let op1 = stack.(sp-2) and op2 = int_of_data (stack.(sp-1)) in
									stack.(sp-2) <- Array(Array.make op2 op1) ; exec fp (sp-1) (pc+1)
	  | Drp    -> exec fp (sp-1) (pc+1)
	  | Bin op -> let op1 = stack.(sp-2) and op2 = stack.(sp-1) in     
	      stack.(sp-2) <- (match (op1, op, op2) with
		| (Integer i1,  Add,     Integer i2)    -> Integer(i1 + i2)
		| (Integer i1,  Sub,     Integer i2)    -> Integer(i1 - i2)
		| (Integer i1,  Mult,    Integer i2)    -> Integer(i1 * i2)
		| (Integer i1,  Div,     Integer i2)    -> Integer(i1 / i2)
		| (Integer i1,  Mod,     Integer i2)    -> Integer(i1 mod i2)
		| (Integer i1,  Equal,   Integer i2)    -> Boolean(i1 = i2)
		| (String i1,   Equal,   String i2)     -> Boolean(i1 = i2)
		| (Boolean i1,  Equal,   Boolean i2)    -> Boolean(i1 = i2)
		| (Integer i1,  Neq,     Integer i2)    -> Boolean(i1 <> i2)
		| (String i1,   Neq,     String i2)     -> Boolean(i1 <> i2)
		| (Boolean i1,  Neq,     Boolean i2)    -> Boolean(i1 <> i2)
		| (Integer i1,  Less,    Integer i2)    -> Boolean(i1 <  i2)
		| (Integer i1,  Leq,     Integer i2)    -> Boolean(i1 <= i2)
		| (Integer i1,  Greater, Integer i2)    -> Boolean(i1 >  i2)
		| (Integer i1,  Geq,     Integer i2)    -> Boolean(i1 >= i2)
		| (Boolean b1,  Or,      Boolean b2)    -> Boolean(b1 || b2)
		| (Boolean b1,  And,     Boolean b2)    -> Boolean(b1 && b2)
		| (Integer i1,  Neg,     Boolean b2)    -> Boolean(not b2)
		| (_,           Concat,  _)     	-> String(string_of_data op1 ^ string_of_data op2)
		| _                                     -> raise (Failure 
								("unexpected binop " ^ string_of_data op1 ^
								" " ^ string_of_stmt (Bin(op)) ^ " " ^ string_of_data op2)));
	      exec fp (sp-1) (pc+1)
	  | Lod i   -> stack.(sp)   <- globals.(i)  ; exec fp (sp+1) (pc+1)
	  | Loda i  -> let ai = int_of_data (stack.(sp-1)) in
										let a = array_of_data (globals.(i)) in
											stack.(sp-1) <- a.(ai) ; exec fp (sp) (pc+1)
	  | Str i   -> require_same_type (globals.(i)) (stack.(sp-1)) ; 
								 globals.(i)  <- stack.(sp-1) ; exec fp sp     (pc+1)
	  | Stra i  -> let ai = int_of_data (stack.(sp-1)) in
										let a = array_of_data (globals.(i)) in
											let e = stack.(sp-2) in
												require_same_type (a.(ai)) (e) ;
												a.(ai) <- e ; exec fp (sp-1) (pc+1)
	  | Lfp i   -> stack.(sp)   <- stack.(fp+i) ; exec fp (sp+1) (pc+1)
	  | Lfpa i  -> let ai = int_of_data (stack.(sp-1)) in 
										let a = array_of_data (stack.(fp+i)) in
											stack.(sp-1) <- a.(ai) ; exec fp (sp) (pc+1)
	  | Sfp i   -> require_same_type (stack.(fp+i)) (stack.(sp-1)) ;
								 stack.(fp+i) <- stack.(sp-1) ; exec fp sp     (pc+1)
	  | Sfpa i  -> let ai = int_of_data (stack.(sp-1)) in  
										let a = array_of_data (stack.(fp+i)) in
											let e = stack.(sp-2) in
												require_same_type (a.(ai)) (e) ;
												a.(ai) <- e ; exec fp (sp-1) (pc+1)
	  | Jsr(-1) -> let prompt = string_of_data (stack.(fp+1)) in 
										print_string (prompt ^ "\n") ; let choices = array_of_data(stack.(fp+2)) in
											(match Array.length choices with
												| 1 ->
														if choices.(0) = String("") then 
															print_string("Enter answer below:\n")
														else
															print_string("Enter one of possible choices below:\n" ^ sstring_of_data choices.(0) ^ "\n")
												| _ -> 
														print_string("Enter one of possible choices below:\n" ^ 
															String.concat "\n" (Array.to_list (Array.map string_of_data choices)) ^ "\n")) ;
											let answers = Array.to_list(array_of_data(stack.(fp+3))) in
												let rsp = String(read_line ()) in
													let rec is_correct ans = function
														| [] -> false
														| hd::tl -> if hd=ans then true else is_correct ans tl in
													let correct_ans = is_correct rsp answers in
														if correct_ans then
															(globals.(0) <- Boolean(true) ;
															globals.(2) <- Integer(int_of_data(globals.(2)) + 1) ;
															print_string("Correct\n"))														 
														else
															(globals.(0) <- Boolean(false) ;
															print_string("Not correct\n")) ;
														globals.(1) <- Integer(int_of_data(globals.(1)) + 1) ;														
										exec fp sp (pc+1)
	  | Jsr(-2) -> let correctCount = int_of_data(globals.(2)) in 
									let askCount = int_of_data(globals.(1)) in
										print_string("Good bye!\n" ^ string_of_int correctCount ^ 
										" out of " ^ string_of_int askCount ^ " answered correctly.\n") ; 
									exec fp sp (pc+1)
	  | Jsr i   -> stack.(sp)   <- Integer(pc + 1)       ; exec fp (sp+1) i
	  | Ent i   -> stack.(sp)   <- Integer(fp) ; clear (sp+1) (sp+i) ; exec sp (sp+i+1) (pc+1)
	  | Rts i   -> let new_fp = int_of_data (stack.(fp)) and new_pc = int_of_data (stack.(fp-1)) in
										stack.(fp-i-1) <- stack.(sp-1) ; exec new_fp (fp-i) new_pc
	  | Beq i   -> exec fp (sp-1) (pc + if stack.(sp-1) =  Boolean(false) then i else 1)
	  | Bne i   -> exec fp (sp-1) (pc + if stack.(sp-1) <> Boolean(false) then i else 1)
	  | Bra i   -> exec fp sp (pc+i)
	  | Hlt     -> ())
	with Failure(s) -> raise(ExecuteFailure("\n" ^ s ^ "\nin: " ^ Bytecode.string_of_stmt (prog.text.(pc))))

  in exec 0 0 0

